#!/usr/bin/env python3
#
# Copyright 2022-2023 The Mumble Developers. All rights reserved.
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file at the root of the
# Mumble source tree or at <https://www.mumble.info/LICENSE>.

from typing import Optional

import argparse
import re
import os

varNameIdx = 0
filePathIdx = 1
displayNameIdx = 2
urlIdx = 3
defineGuardIdx = 4

repoRoot = os.path.normpath(os.path.join(os.path.dirname(__file__), ".."))

# List of (3rdparty) licenses  [<variableName>, <pathToLicenseFile>, <DisplayName>, <URL>, (<define-guard>)?]
licenses = [
    ["licenseMumble", "LICENSE", "Mumble", "https://www.mumble.info/"],
    ["licenseEmojiOne", "icons/flags/LICENSE.md", "Emoji One artwork", "http://emojione.com/"],
    ["licenseQQBonjour", "3rdparty/qqbonjour/LICENSE", "QQBonjour", "https://doc.qt.io/archives/qq/qq23-bonjour.html"],
    ["licenseSmallFT", "3rdparty/smallft/LICENSE", "smallft", "https://www.xiph.org"],
    ["licenseArc4RandomUniform", "3rdparty/arc4random/LICENSE", "arc4random_uniform", "https://www.openbsd.org"],
]


def formatLicenseVar(var_name: str, license_path: str, define_guard: Optional[str]) -> str:
    with open(os.path.join(repoRoot, license_path), "r") as file:
        content = file.read().strip().replace("\"", "\\\"")

        content = "\"" + re.sub("\\n\\s*", "\\\\n\"\n\t\"", content) + "\""

        content = "static const char *" + var_name + " = " + content + ";"

        if define_guard:
            content = "#ifdef " + define_guard + "\n" + content + "\n" + "#endif // " + define_guard + "\n"

        return content


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--output", help="Path to the file to which the generated code shall be written." +
                        " If none is given, the content is written to stdout", metavar="PATH")

    args = parser.parse_args()

    content = """/*
 * This file was auto-generated by scripts/mklic.pl\n
 * DO NOT EDIT IT MANUALLY
 */
#ifndef MUMBLE_LICENSES_H_
#define MUMBLE_LICENSES_H_

struct ThirdPartyLicense {
	const char* name;
	const char* url;
	const char* license;

	ThirdPartyLicense() : name(nullptr), url(nullptr), license(nullptr) {}
	ThirdPartyLicense(const char* name_, const char* url_, const char* license_)
	    : name(name_), url(url_), license(license_) {}

	bool isEmpty() const { return !name && !url && !license; }
};
"""
    
    for currentLicense in licenses:
        content += "\n" + formatLicenseVar(var_name=currentLicense[varNameIdx], license_path=currentLicense[filePathIdx],
                define_guard=currentLicense[defineGuardIdx] if len(currentLicense) > defineGuardIdx else None)


    content += "\n\n"
    content += "static const ThirdPartyLicense licenses3rdParties[] = {\n"

    for currentLicense in licenses[ 1 : ]:
        if len(currentLicense) > defineGuardIdx:
            content += "#ifdef " + currentLicense[defineGuardIdx] + "\n"

        content += "\t" + ("ThirdPartyLicense(\"%s\", \"%s\", %s)" % (currentLicense[displayNameIdx], currentLicense[urlIdx],
            currentLicense[varNameIdx])) + ",\n"

        if len(currentLicense) > defineGuardIdx:
            content += "#endif //" + currentLicense[defineGuardIdx] + "\n"

    content += "\t// Empty entry that marks the end of the array\n"
    content += "\tThirdPartyLicense()\n"
    content += "};\n"
    content += "#endif // MUMBLE_LICENSES_H_\n"

    if args.output:
        with open(args.output, "w") as outFile:
            outFile.write(content)
    else:
        print(content)


if __name__ == "__main__":
    main()
